@@ -1,6 +1,7 @@
 from sklearn.model_selection import GridSearchCV
 from sklearn.datasets import load_boston
 import xgboost as xgb
+import multiprocessing
 
 if __name__ == "__main__":
     print("Parallel Parameter optimization")
@@ -8,7 +9,7 @@ if __name__ == "__main__":
 
     y = boston['target']
     X = boston['data']
-    xgb_model = xgb.XGBRegressor()
+    xgb_model = xgb.XGBRegressor(n_jobs=multiprocessing.cpu_count() // 2)
     clf = GridSearchCV(xgb_model, {'max_depth': [2, 4, 6],
                                    'n_estimators': [50, 100, 200]}, verbose=1,
                        n_jobs=2)
