@@ -1,5 +1,5 @@
 /*!
- * Copyright 2019 by Contributors
+ * Copyright 2019-2021 by Contributors
  */
 #include <utility>
 #include <limits>
@@ -10,10 +10,6 @@ namespace xgboost {
 namespace gbm {
 
 void GBLinearModel::SaveModel(Json* p_out) const {
-  using WeightType = std::remove_reference<decltype(std::declval<decltype(weight)>().back())>::type;
-  using JsonFloat = Number::Float;
-  static_assert(std::is_same<WeightType, JsonFloat>::value,
-                "Weight type should be of the same type with JSON float");
   auto& out = *p_out;
 
   size_t const n_weights = weight.size();
@@ -22,6 +18,7 @@ void GBLinearModel::SaveModel(Json* p_out) const {
     j_weights[i] = weight[i];
   }
   out["weights"] = std::move(j_weights);
+  out["boosted_rounds"] = Json{this->num_boosted_rounds};
 }
 
 void GBLinearModel::LoadModel(Json const& in) {
@@ -31,6 +28,13 @@ void GBLinearModel::LoadModel(Json const& in) {
   for (size_t i = 0; i < n_weights; ++i) {
     weight[i] = get<Number const>(j_weights[i]);
   }
+  auto const& obj = get<Object const>(in);
+  auto boosted_rounds = obj.find("boosted_rounds");
+  if (boosted_rounds != obj.cend()) {
+    this->num_boosted_rounds = get<Integer const>(boosted_rounds->second);
+  } else {
+    this->num_boosted_rounds = 0;
+  }
 }
 
 DMLC_REGISTER_PARAMETER(DeprecatedGBLinearModelParam);
