@@ -14,6 +14,7 @@
 #include "./param.h"
 #include "../gbm/gblinear_model.h"
 #include "../common/random.h"
+#include "../common/threading_utils.h"
 
 namespace xgboost {
 namespace linear {
@@ -82,7 +83,8 @@ inline std::pair<double, double> GetGradient(int group_idx, int num_group, int f
                                              DMatrix *p_fmat) {
   double sum_grad = 0.0, sum_hess = 0.0;
   for (const auto &batch : p_fmat->GetBatches<CSCPage>()) {
-    auto col = batch[fidx];
+    auto page = batch.GetView();
+    auto col = page[fidx];
     const auto ndata = static_cast<bst_omp_uint>(col.size());
     for (bst_omp_uint j = 0; j < ndata; ++j) {
       const bst_float v = col[j].fvalue;
@@ -111,16 +113,21 @@ inline std::pair<double, double> GetGradientParallel(int group_idx, int num_grou
                                                      DMatrix *p_fmat) {
   double sum_grad = 0.0, sum_hess = 0.0;
   for (const auto &batch : p_fmat->GetBatches<CSCPage>()) {
-    auto col = batch[fidx];
+    auto page = batch.GetView();
+    auto col = page[fidx];
     const auto ndata = static_cast<bst_omp_uint>(col.size());
+    dmlc::OMPException exc;
 #pragma omp parallel for schedule(static) reduction(+ : sum_grad, sum_hess)
     for (bst_omp_uint j = 0; j < ndata; ++j) {
-      const bst_float v = col[j].fvalue;
-      auto &p = gpair[col[j].index * num_group + group_idx];
-      if (p.GetHess() < 0.0f) continue;
-      sum_grad += p.GetGrad() * v;
-      sum_hess += p.GetHess() * v * v;
+      exc.Run([&]() {
+        const bst_float v = col[j].fvalue;
+        auto &p = gpair[col[j].index * num_group + group_idx];
+        if (p.GetHess() < 0.0f) return;
+        sum_grad += p.GetGrad() * v;
+        sum_hess += p.GetHess() * v * v;
+      });
     }
+    exc.Rethrow();
   }
   return std::make_pair(sum_grad, sum_hess);
 }
@@ -140,14 +147,18 @@ inline std::pair<double, double> GetBiasGradientParallel(int group_idx, int num_
                                                          DMatrix *p_fmat) {
   double sum_grad = 0.0, sum_hess = 0.0;
   const auto ndata = static_cast<bst_omp_uint>(p_fmat->Info().num_row_);
+  dmlc::OMPException exc;
 #pragma omp parallel for schedule(static) reduction(+ : sum_grad, sum_hess)
   for (bst_omp_uint i = 0; i < ndata; ++i) {
-    auto &p = gpair[i * num_group + group_idx];
-    if (p.GetHess() >= 0.0f) {
-      sum_grad += p.GetGrad();
-      sum_hess += p.GetHess();
-    }
+    exc.Run([&]() {
+      auto &p = gpair[i * num_group + group_idx];
+      if (p.GetHess() >= 0.0f) {
+        sum_grad += p.GetGrad();
+        sum_hess += p.GetHess();
+      }
+    });
   }
+  exc.Rethrow();
   return std::make_pair(sum_grad, sum_hess);
 }
 
@@ -166,15 +177,20 @@ inline void UpdateResidualParallel(int fidx, int group_idx, int num_group,
                                    DMatrix *p_fmat) {
   if (dw == 0.0f) return;
   for (const auto &batch : p_fmat->GetBatches<CSCPage>()) {
-    auto col = batch[fidx];
+    auto page = batch.GetView();
+    auto col = page[fidx];
     // update grad value
     const auto num_row = static_cast<bst_omp_uint>(col.size());
+    dmlc::OMPException exc;
 #pragma omp parallel for schedule(static)
     for (bst_omp_uint j = 0; j < num_row; ++j) {
-      GradientPair &p = (*in_gpair)[col[j].index * num_group + group_idx];
-      if (p.GetHess() < 0.0f) continue;
-      p += GradientPair(p.GetHess() * col[j].fvalue * dw, 0);
+      exc.Run([&]() {
+        GradientPair &p = (*in_gpair)[col[j].index * num_group + group_idx];
+        if (p.GetHess() < 0.0f) return;
+        p += GradientPair(p.GetHess() * col[j].fvalue * dw, 0);
+      });
     }
+    exc.Rethrow();
   }
 }
 
@@ -192,12 +208,16 @@ inline void UpdateBiasResidualParallel(int group_idx, int num_group, float dbias
                                        DMatrix *p_fmat) {
   if (dbias == 0.0f) return;
   const auto ndata = static_cast<bst_omp_uint>(p_fmat->Info().num_row_);
+  dmlc::OMPException exc;
 #pragma omp parallel for schedule(static)
   for (bst_omp_uint i = 0; i < ndata; ++i) {
-    GradientPair &g = (*in_gpair)[i * num_group + group_idx];
-    if (g.GetHess() < 0.0f) continue;
-    g += GradientPair(g.GetHess() * dbias, 0);
+    exc.Run([&]() {
+      GradientPair &g = (*in_gpair)[i * num_group + group_idx];
+      if (g.GetHess() < 0.0f) return;
+      g += GradientPair(g.GetHess() * dbias, 0);
+    });
   }
+  exc.Rethrow();
 }
 
 /**
@@ -220,10 +240,10 @@ class FeatureSelector {
    * \param lambda Regularisation lambda.
    * \param param  A parameter with algorithm-dependent use.
    */
-  virtual void Setup(const gbm::GBLinearModel &model,
-                     const std::vector<GradientPair> &gpair,
-                     DMatrix *p_fmat,
-                     float alpha, float lambda, int param) {}
+  virtual void Setup(const gbm::GBLinearModel &,
+                     const std::vector<GradientPair> &,
+                     DMatrix *,
+                     float , float , int ) {}
   /**
    * \brief Select next coordinate to update.
    *
@@ -250,8 +270,8 @@ class FeatureSelector {
 class CyclicFeatureSelector : public FeatureSelector {
  public:
   int NextFeature(int iteration, const gbm::GBLinearModel &model,
-                  int group_idx, const std::vector<GradientPair> &gpair,
-                  DMatrix *p_fmat, float alpha, float lambda) override {
+                  int , const std::vector<GradientPair> &,
+                  DMatrix *, float, float) override {
     return iteration % model.learner_model_param->num_feature;
   }
 };
@@ -263,8 +283,8 @@ class CyclicFeatureSelector : public FeatureSelector {
 class ShuffleFeatureSelector : public FeatureSelector {
  public:
   void Setup(const gbm::GBLinearModel &model,
-             const std::vector<GradientPair> &gpair,
-             DMatrix *p_fmat, float alpha, float lambda, int param) override {
+             const std::vector<GradientPair>&,
+             DMatrix *, float, float, int) override {
     if (feat_index_.size() == 0) {
       feat_index_.resize(model.learner_model_param->num_feature);
       std::iota(feat_index_.begin(), feat_index_.end(), 0);
@@ -273,8 +293,8 @@ class ShuffleFeatureSelector : public FeatureSelector {
   }
 
   int NextFeature(int iteration, const gbm::GBLinearModel &model,
-                  int group_idx, const std::vector<GradientPair> &gpair,
-                  DMatrix *p_fmat, float alpha, float lambda) override {
+                  int, const std::vector<GradientPair> &,
+                  DMatrix *, float, float) override {
     return feat_index_[iteration % model.learner_model_param->num_feature];
   }
 
@@ -288,9 +308,9 @@ class ShuffleFeatureSelector : public FeatureSelector {
  */
 class RandomFeatureSelector : public FeatureSelector {
  public:
-  int NextFeature(int iteration, const gbm::GBLinearModel &model,
-                  int group_idx, const std::vector<GradientPair> &gpair,
-                  DMatrix *p_fmat, float alpha, float lambda) override {
+  int NextFeature(int, const gbm::GBLinearModel &model,
+                  int, const std::vector<GradientPair> &,
+                  DMatrix *, float, float) override {
     return common::GlobalRandom()() % model.learner_model_param->num_feature;
   }
 };
@@ -307,8 +327,8 @@ class RandomFeatureSelector : public FeatureSelector {
 class GreedyFeatureSelector : public FeatureSelector {
  public:
   void Setup(const gbm::GBLinearModel &model,
-             const std::vector<GradientPair> &gpair,
-             DMatrix *p_fmat, float alpha, float lambda, int param) override {
+             const std::vector<GradientPair> &,
+             DMatrix *, float, float, int param) override {
     top_k_ = static_cast<bst_uint>(param);
     const bst_uint ngroup = model.learner_model_param->num_output_group;
     if (param <= 0) top_k_ = std::numeric_limits<bst_uint>::max();
@@ -321,7 +341,7 @@ class GreedyFeatureSelector : public FeatureSelector {
     }
   }
 
-  int NextFeature(int iteration, const gbm::GBLinearModel &model,
+  int NextFeature(int, const gbm::GBLinearModel &model,
                   int group_idx, const std::vector<GradientPair> &gpair,
                   DMatrix *p_fmat, float alpha, float lambda) override {
     // k-th selected feature for a group
@@ -333,10 +353,10 @@ class GreedyFeatureSelector : public FeatureSelector {
     const bst_omp_uint nfeat = model.learner_model_param->num_feature;
     // Calculate univariate gradient sums
     std::fill(gpair_sums_.begin(), gpair_sums_.end(), std::make_pair(0., 0.));
-  for (const auto &batch : p_fmat->GetBatches<CSCPage>()) {
-      #pragma omp parallel for schedule(static)
-      for (bst_omp_uint i = 0; i < nfeat; ++i) {
-        const auto col = batch[i];
+    for (const auto &batch : p_fmat->GetBatches<CSCPage>()) {
+      auto page = batch.GetView();
+      common::ParallelFor(nfeat, [&](bst_omp_uint i) {
+        const auto col = page[i];
         const bst_uint ndata = col.size();
         auto &sums = gpair_sums_[group_idx * nfeat + i];
         for (bst_uint j = 0u; j < ndata; ++j) {
@@ -346,7 +366,7 @@ class GreedyFeatureSelector : public FeatureSelector {
           sums.first += p.GetGrad() * v;
           sums.second += p.GetHess() * v * v;
         }
-      }
+      });
     }
     // Find a feature with the largest magnitude of weight change
     int best_fidx = 0;
@@ -399,10 +419,10 @@ class ThriftyFeatureSelector : public FeatureSelector {
     // Calculate univariate gradient sums
     std::fill(gpair_sums_.begin(), gpair_sums_.end(), std::make_pair(0., 0.));
     for (const auto &batch : p_fmat->GetBatches<CSCPage>()) {
-// column-parallel is usually faster than row-parallel
-#pragma omp parallel for schedule(static)
-      for (bst_omp_uint i = 0; i < nfeat; ++i) {
-        const auto col = batch[i];
+      auto page = batch.GetView();
+      // column-parallel is usually fastaer than row-parallel
+      common::ParallelFor(nfeat, [&](bst_omp_uint i) {
+        const auto col = page[i];
         const bst_uint ndata = col.size();
         for (bst_uint gid = 0u; gid < ngroup; ++gid) {
           auto &sums = gpair_sums_[gid * nfeat + i];
@@ -414,7 +434,7 @@ class ThriftyFeatureSelector : public FeatureSelector {
             sums.second += p.GetHess() * v * v;
           }
         }
-      }
+      });
     }
     // rank by descending weight magnitude within the groups
     std::fill(deltaw_.begin(), deltaw_.end(), 0.f);
@@ -438,9 +458,9 @@ class ThriftyFeatureSelector : public FeatureSelector {
     }
   }
 
-  int NextFeature(int iteration, const gbm::GBLinearModel &model,
-                  int group_idx, const std::vector<GradientPair> &gpair,
-                  DMatrix *p_fmat, float alpha, float lambda) override {
+  int NextFeature(int, const gbm::GBLinearModel &model,
+                  int group_idx, const std::vector<GradientPair> &,
+                  DMatrix *, float, float) override {
     // k-th selected feature for a group
     auto k = counter_[group_idx]++;
     // stop after either reaching top-N or going through all the features in a group
