@@ -56,13 +56,6 @@ class HistMaker: public BaseMaker {
     HistUnit(const float *cut, GradStats *data, uint32_t size)
         : cut{cut}, data{data}, size{size} {}
     /*! \brief add a histogram to data */
-    void Add(float fv, const std::vector<GradientPair> &gpair,
-             const MetaInfo &info, const size_t ridx) {
-      unsigned bin = std::upper_bound(cut, cut + size, fv) - cut;
-      CHECK_NE(size, 0U) << "try insert into size=0";
-      CHECK_LT(bin, size);
-      data[bin].Add(gpair[ridx]);
-    }
   };
   /*! \brief a set of histograms from different index */
   struct HistSet {
@@ -86,7 +79,7 @@ class HistMaker: public BaseMaker {
     // per thread histset
     std::vector<HistSet> hset;
     // initialize the hist set
-    inline void Configure(const TrainParam &param, int nthread) {
+    inline void Configure(int nthread) {
       hset.resize(nthread);
       // cleanup statistics
       for (int tid = 0; tid < nthread; ++tid) {
@@ -127,7 +120,7 @@ class HistMaker: public BaseMaker {
       // create histogram
       this->CreateHist(gpair, p_fmat, selected_features_, *p_tree);
       // find split based on histogram statistics
-      this->FindSplit(depth, gpair, p_fmat, selected_features_, p_tree);
+      this->FindSplit(selected_features_, p_tree);
       // reset position after split
       this->ResetPositionAfterSplit(p_fmat, *p_tree);
       this->UpdateQueueExpand(*p_tree);
@@ -146,7 +139,7 @@ class HistMaker: public BaseMaker {
                                   const std::vector <bst_feature_t> &fset,
                                   const RegTree &tree) = 0;
   // initialize the current working set of features in this round
-  virtual void InitWorkSet(DMatrix *p_fmat,
+  virtual void InitWorkSet(DMatrix *,
                            const RegTree &tree,
                            std::vector<bst_feature_t> *p_fset) {
     p_fset->resize(tree.param.num_feature);
@@ -159,9 +152,9 @@ class HistMaker: public BaseMaker {
                                        const RegTree &tree) {
   }
   virtual void CreateHist(const std::vector<GradientPair> &gpair,
-                          DMatrix *p_fmat,
+                          DMatrix *,
                           const std::vector <bst_feature_t> &fset,
-                          const RegTree &tree)  = 0;
+                          const RegTree &)  = 0;
 
  private:
   void EnumerateSplit(const HistUnit &hist,
@@ -202,32 +195,33 @@ class HistMaker: public BaseMaker {
     }
   }
 
-  void FindSplit(int depth,
-                 const std::vector<GradientPair> &gpair,
-                 DMatrix *p_fmat,
-                 const std::vector <bst_feature_t> &feature_set,
+  void FindSplit(const std::vector <bst_feature_t> &feature_set,
                  RegTree *p_tree) {
     const size_t num_feature = feature_set.size();
     // get the best split condition for each node
     std::vector<SplitEntry> sol(qexpand_.size());
     std::vector<GradStats> left_sum(qexpand_.size());
     auto nexpand = static_cast<bst_omp_uint>(qexpand_.size());
+    dmlc::OMPException exc;
 #pragma omp parallel for schedule(dynamic, 1)
     for (bst_omp_uint wid = 0; wid < nexpand; ++wid) {
-      const int nid = qexpand_[wid];
-      CHECK_EQ(node2workindex_[nid], static_cast<int>(wid));
-      SplitEntry &best = sol[wid];
-      GradStats &node_sum = wspace_.hset[0][num_feature + wid * (num_feature + 1)].data[0];
-      for (size_t i = 0; i < feature_set.size(); ++i) {
-        // Query is thread safe as it's a const function.
-        if (!this->interaction_constraints_.Query(nid, feature_set[i])) {
-          continue;
-        }
+      exc.Run([&]() {
+        const int nid = qexpand_[wid];
+        CHECK_EQ(node2workindex_[nid], static_cast<int>(wid));
+        SplitEntry &best = sol[wid];
+        GradStats &node_sum = wspace_.hset[0][num_feature + wid * (num_feature + 1)].data[0];
+        for (size_t i = 0; i < feature_set.size(); ++i) {
+          // Query is thread safe as it's a const function.
+          if (!this->interaction_constraints_.Query(nid, feature_set[i])) {
+            continue;
+          }
 
-        EnumerateSplit(this->wspace_.hset[0][i + wid * (num_feature+1)],
-                       node_sum, feature_set[i], &best, &left_sum[wid]);
-      }
+          EnumerateSplit(this->wspace_.hset[0][i + wid * (num_feature+1)],
+                        node_sum, feature_set[i], &best, &left_sum[wid]);
+        }
+      });
     }
+    exc.Rethrow();
     // get the best result, we can synchronize the solution
     for (bst_omp_uint wid = 0; wid < nexpand; ++wid) {
       const bst_node_t nid = qexpand_[wid];
@@ -288,7 +282,6 @@ class CQHistMaker: public HistMaker {
      */
     inline void Add(bst_float fv,
                     const std::vector<GradientPair> &gpair,
-                    const MetaInfo &info,
                     const bst_uint ridx) {
       while (istart < hist.size && !(fv < hist.cut[istart])) ++istart;
       CHECK_NE(istart, hist.size);
@@ -342,25 +335,30 @@ class CQHistMaker: public HistMaker {
       feat2workindex_[fset[i]] = static_cast<int>(i);
     }
     // start to work
-    this->wspace_.Configure(this->param_, 1);
+    this->wspace_.Configure(1);
     // if it is C++11, use lazy evaluation for Allreduce,
     // to gain speedup in recovery
     auto lazy_get_hist = [&]() {
       thread_hist_.resize(omp_get_max_threads());
       // start accumulating statistics
       for (const auto &batch : p_fmat->GetBatches<SortedCSCPage>()) {
+        auto page = batch.GetView();
         // start enumeration
         const auto nsize = static_cast<bst_omp_uint>(fset.size());
+        dmlc::OMPException exc;
 #pragma omp parallel for schedule(dynamic, 1)
         for (bst_omp_uint i = 0; i < nsize; ++i) {
-          int fid = fset[i];
-          int offset = feat2workindex_[fid];
-          if (offset >= 0) {
-            this->UpdateHistCol(gpair, batch[fid], info, tree,
-                                fset, offset,
-                                &thread_hist_[omp_get_thread_num()]);
-          }
+          exc.Run([&]() {
+            int fid = fset[i];
+            int offset = feat2workindex_[fid];
+            if (offset >= 0) {
+              this->UpdateHistCol(gpair, page[fid], info, tree,
+                                  fset, offset,
+                                  &thread_hist_[omp_get_thread_num()]);
+            }
+          });
         }
+        exc.Rethrow();
       }
       // update node statistics.
       this->GetNodeStats(gpair, *p_fmat, tree,
@@ -376,7 +374,7 @@ class CQHistMaker: public HistMaker {
                              this->wspace_.hset[0].data.size(), lazy_get_hist);
   }
 
-  void ResetPositionAfterSplit(DMatrix *p_fmat,
+  void ResetPositionAfterSplit(DMatrix *,
                                  const RegTree &tree) override {
     this->GetSplitSet(this->qexpand_, tree, &fsplit_set_);
   }
@@ -403,7 +401,7 @@ class CQHistMaker: public HistMaker {
     for (auto& sketch : sketchs_) {
       sketch.Init(info.num_row_, this->param_.sketch_eps);
     }
-    // intitialize the summary array
+    // initialize the summary array
     summary_array_.resize(sketchs_.size());
     // setup maximum size
     unsigned max_size = this->param_.MaxSketchSize();
@@ -411,7 +409,7 @@ class CQHistMaker: public HistMaker {
       summary_array_[i].Reserve(max_size);
     }
     {
-      // get smmary
+      // get summary
       thread_sketch_.resize(omp_get_max_threads());
 
       // TWOPASS: use the real set + split set in the column iteration.
@@ -424,19 +422,23 @@ class CQHistMaker: public HistMaker {
       for (const auto &batch : p_fmat->GetBatches<SortedCSCPage>()) {
         // TWOPASS: use the real set + split set in the column iteration.
         this->CorrectNonDefaultPositionByBatch(batch, fsplit_set_, tree);
-
+        auto page = batch.GetView();
         // start enumeration
         const auto nsize = static_cast<bst_omp_uint>(work_set_.size());
-        #pragma omp parallel for schedule(dynamic, 1)
+        dmlc::OMPException exc;
+#pragma omp parallel for schedule(dynamic, 1)
         for (bst_omp_uint i = 0; i < nsize; ++i) {
-          int fid = work_set_[i];
-          int offset = feat2workindex_[fid];
-          if (offset >= 0) {
-            this->UpdateSketchCol(gpair, batch[fid], tree,
-                                  work_set_size, offset,
-                                  &thread_sketch_[omp_get_thread_num()]);
-          }
+          exc.Run([&]() {
+            int fid = work_set_[i];
+            int offset = feat2workindex_[fid];
+            if (offset >= 0) {
+              this->UpdateSketchCol(gpair, page[fid], tree,
+                                    work_set_size, offset,
+                                    &thread_sketch_[omp_get_thread_num()]);
+            }
+          });
         }
+        exc.Rethrow();
       }
       for (size_t i = 0; i < sketchs_.size(); ++i) {
         common::WXQuantileSketch<bst_float, bst_float>::SummaryContainer out;
@@ -533,7 +535,7 @@ class CQHistMaker: public HistMaker {
         const bst_uint ridx = c.index;
         const int nid = this->position_[ridx];
         if (nid >= 0) {
-          hbuilder[nid].Add(c.fvalue, gpair, info, ridx);
+          hbuilder[nid].Add(c.fvalue, gpair, ridx);
         }
       }
     }
@@ -689,7 +691,7 @@ class GlobalProposalHistMaker: public CQHistMaker {
       this->feat2workindex_[fset[i]] = static_cast<int>(i);
     }
     // start to work
-    this->wspace_.Configure(this->param_, 1);
+    this->wspace_.Configure(1);
     // to gain speedup in recovery
     {
       this->thread_hist_.resize(omp_get_max_threads());
@@ -707,19 +709,24 @@ class GlobalProposalHistMaker: public CQHistMaker {
       for (const auto &batch : p_fmat->GetBatches<SortedCSCPage>()) {
         // TWOPASS: use the real set + split set in the column iteration.
         this->CorrectNonDefaultPositionByBatch(batch, this->fsplit_set_, tree);
+        auto page = batch.GetView();
 
         // start enumeration
         const auto nsize = static_cast<bst_omp_uint>(this->work_set_.size());
+        dmlc::OMPException exc;
 #pragma omp parallel for schedule(dynamic, 1)
         for (bst_omp_uint i = 0; i < nsize; ++i) {
-          int fid = this->work_set_[i];
-          int offset = this->feat2workindex_[fid];
-          if (offset >= 0) {
-            this->UpdateHistCol(gpair, batch[fid], info, tree,
-                                fset, offset,
-                                &this->thread_hist_[omp_get_thread_num()]);
-          }
+          exc.Run([&]() {
+            int fid = this->work_set_[i];
+            int offset = this->feat2workindex_[fid];
+            if (offset >= 0) {
+              this->UpdateHistCol(gpair, page[fid], info, tree,
+                                  fset, offset,
+                                  &this->thread_hist_[omp_get_thread_num()]);
+            }
+          });
         }
+        exc.Rethrow();
       }
 
       // update node statistics.
