@@ -2,7 +2,6 @@ import pickle
 import numpy as np
 import xgboost as xgb
 import os
-import unittest
 
 
 kRows = 100
@@ -15,7 +14,7 @@ def generate_data():
     return X, y
 
 
-class TestPickling(unittest.TestCase):
+class TestPickling:
     def run_model_pickling(self, xgb_params):
         X, y = generate_data()
         dtrain = xgb.DMatrix(X, y)
@@ -43,17 +42,9 @@ class TestPickling(unittest.TestCase):
         if os.path.exists(filename):
             os.remove(filename)
 
-    def test_model_pickling_binary(self):
-        params = {
-            'nthread': 1,
-            'tree_method': 'hist'
-        }
-        self.run_model_pickling(params)
-
     def test_model_pickling_json(self):
         params = {
             'nthread': 1,
             'tree_method': 'hist',
-            'enable_experimental_json_serialization': True
         }
         self.run_model_pickling(params)
